#!/usr/bin/env python
# -*- coding: utf-8 -*

"""
    Class for the testing suite :
    - get the list of all test files
    - create a copy of them on start
    - remove the copy on end
"""

import shutil
import os
import glob
import sys
import tempfile
import unittest
import subprocess

VERBOSITY = 15

clean = glob.glob('clean*')
clean.sort()
dirty = glob.glob('dirty*')
dirty.sort()

FILE_LIST = zip(clean, dirty)

try:  # PDF render processing
    import cairo
    import gi

    gi.require_version('Poppler', '0.18')
    from gi.repository import Poppler
    import pdfrw
except ImportError:
    FILE_LIST.remove(('clean é.pdf', 'dirty é.pdf'))

try:  # python-mutagen : audio file format
    import mutagen
except ImportError:
    FILE_LIST.remove(('clean é.ogg', 'dirty é.ogg'))
    FILE_LIST.remove(('clean é.mp3', 'dirty é.mp3'))
    FILE_LIST.remove(('clean é.flac', 'dirty é.flac'))

try:  # exiftool
    subprocess.check_output(['exiftool', '-ver'])
except OSError:
    FILE_LIST.remove(('clean é.tif', 'dirty é.tif'))


class MATTest(unittest.TestCase):
    """
        Parent class of all test-functions
    """

    def setUp(self):
        """
            Create working copy of the clean and the dirty file in the TMP dir
        """
        self.file_list = []
        self.tmpdir = tempfile.mkdtemp()

        for clean_file, dirty_file in FILE_LIST:
            clean_dir = os.path.join(self.tmpdir, clean_file)
            dirty_dir = os.path.join(self.tmpdir, dirty_file)
            shutil.copy2(clean_file, clean_dir)
            shutil.copy2(dirty_file, dirty_dir)
            self.file_list.append((clean_dir, dirty_dir))

    def tearDown(self):
        """
            Remove the tmp folder
        """
        for root, dirs, files in os.walk(self.tmpdir):
            for d in dirs + files:
                os.chmod(os.path.join(root, d), 0o777)
        shutil.rmtree(self.tmpdir)


def run_all_tests():
    """
    This method will run all tests, both for cli and lib.
    The imports of clitest and libtest are done here because
    we're modifying the PATH (technically, it's two path:
    the one used to spawn the `mat` process, and the one for Python import)
    in the main function.
    """
    import clitest
    import libtest
    suite = unittest.TestSuite()
    suite.addTests(clitest.get_tests())
    suite.addTests(libtest.get_tests())

    return unittest.TextTestRunner(verbosity=VERBOSITY).run(suite).wasSuccessful()


def set_local():
    """ Monkey patch paths to run the testsuite on the _local_
    version of MAT. See `run_all_tests` for more information about
    what paths we're changing and why.
    """
    os.environ['PATH'] = '..:' + os.environ['PATH']
    sys.path.append('..')


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(description='MAT testsuite')
    parser.add_argument('-s', '--system', action='store_true',
                        help='Test the system-wide version of mat')

    if parser.parse_args().system is False:
        set_local()

    sys.exit(run_all_tests() is False)



""" Care about office's formats

"""

import logging
import os
import shutil
import tempfile
import xml.dom.minidom as minidom
import zipfile

try:
    import cairo
    import gi
    gi.require_version('Poppler', '0.18')
    from gi.repository import Poppler
except ImportError:
    logging.info('office.py loaded without PDF support')

import parser
import archive


class OpenDocumentStripper(archive.TerminalZipStripper):
    """ An open document file is a zip, with xml file into.
        The one that interest us is meta.xml
    """

    def get_meta(self):
        """ Return a dict with all the meta of the file by
            trying to read the meta.xml file.
        """
        metadata = super(OpenDocumentStripper, self).get_meta()
        zipin = zipfile.ZipFile(self.filename, 'r')
        try:
            content = zipin.read('meta.xml')
            dom1 = minidom.parseString(content)
            elements = dom1.getElementsByTagName('office:meta')
            for i in elements[0].childNodes:
                if i.tagName != 'meta:document-statistic':
                    nodename = ''.join(i.nodeName.split(':')[1:])
                    metadata[nodename] = ''.join([j.data for j in i.childNodes])
        except KeyError:  # no meta.xml file found
            logging.debug('%s has no opendocument metadata', self.filename)
        zipin.close()
        return metadata

    def remove_all(self):
        """ Removes metadata
        """
        return super(OpenDocumentStripper, self).remove_all(ending_blacklist=['meta.xml'])

    def is_clean(self):
        """ Check if the file is clean from harmful metadatas
        """
        clean_super = super(OpenDocumentStripper, self).is_clean()
        if clean_super is False:
            return False

        zipin = zipfile.ZipFile(self.filename, 'r')
        try:
            zipin.getinfo('meta.xml')
        except KeyError:  # no meta.xml in the file
            return True
        zipin.close()
        return False


class OpenXmlStripper(archive.TerminalZipStripper):
    """ Represent an office openxml document, which is like
        an opendocument format, with some tricky stuff added.
        It contains mostly xml, but can have media blobs, crap, ...
        (I don't like this format.)
    """

    def remove_all(self):
        """ Remove harmful metadata, by deleting everything that doesn't end with '.rels' in the
        'docProps' folder. """
        return super(OpenXmlStripper, self).remove_all(
            beginning_blacklist=['docProps/'], whitelist=['.rels'])

    def is_clean(self):
        """ Check if the file is clean from harmful metadatas.
            This implementation is faster than something like
            "return this.get_meta() == {}".
        """
        clean_super = super(OpenXmlStripper, self).is_clean()
        if clean_super is False:
            return False

        zipin = zipfile.ZipFile(self.filename)
        for item in zipin.namelist():
            if item.startswith('docProps/'):
                return False
        zipin.close()
        return True

    def get_meta(self):
        """ Return a dict with all the meta of the file
        """
        metadata = super(OpenXmlStripper, self).get_meta()

        zipin = zipfile.ZipFile(self.filename)
        for item in zipin.namelist():
            if item.startswith('docProps/'):
                metadata[item] = 'harmful content'
        zipin.close()
        return metadata


class PdfStripper(parser.GenericParser):
    """ Represent a PDF file
    """

    def __init__(self, filename, mime, backup, is_writable, **kwargs):
        super(PdfStripper, self).__init__(filename, mime, backup, is_writable, **kwargs)
        self.uri = 'file://' + os.path.abspath(self.filename)
        self.password = None
        try:
            self.pdf_quality = kwargs['low_pdf_quality']
        except KeyError:
            self.pdf_quality = False

        self.meta_list = frozenset(['title', 'author', 'subject',
                                    'keywords', 'creator', 'producer', 'metadata'])

    def is_clean(self):
        """ Check if the file is clean from harmful metadatas
        """
        document = Poppler.Document.new_from_file(self.uri, self.password)
        return not any(document.get_property(key) for key in self.meta_list)

    def remove_all(self):
        """ Opening the PDF with poppler, then doing a render
            on a cairo pdfsurface for each pages.

            http://cairographics.org/documentation/pycairo/2/

            The use of an intermediate tempfile is necessary because
            python-cairo segfaults on unicode.
            See http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=699457
        """
        document = Poppler.Document.new_from_file(self.uri, self.password)
        try:
            output = tempfile.mkstemp()[1]

            # Size doesn't matter (pun intended),
            # since the surface will be resized before
            # being rendered
            surface = cairo.PDFSurface(output, 10, 10)
            context = cairo.Context(surface)  # context draws on the surface

            logging.debug('PDF rendering of %s', self.filename)
            for pagenum in range(document.get_n_pages()):
                page = document.get_page(pagenum)
                page_width, page_height = page.get_size()
                surface.set_size(page_width, page_height)
                context.save()
                if self.pdf_quality:  # this may reduce the produced PDF size
                    page.render(context)
                else:
                    page.render_for_printing(context)
                context.restore()
                context.show_page()  # draw context on surface
            surface.finish()
            shutil.move(output, self.output)
        except:
            logging.error('Something went wrong when cleaning %s.', self.filename)
            return False

        try:
            # For now, cairo cannot write meta, so we must use pdfrw
            # See the realted thread: http://lists.cairographics.org/archives/cairo/2007-September/011466.html
            import pdfrw

            logging.debug('Removing %s\'s superficial metadata', self.filename)
            trailer = pdfrw.PdfReader(self.output)
            trailer.Info.Producer = None
            trailer.Info.Creator = None
            writer = pdfrw.PdfWriter()
            writer.trailer = trailer
            writer.write(self.output)
            self.do_backup()
        except:
            logging.error('Unable to remove all metadata from %s, please install pdfrw', self.output)
            return False
        return True

    def get_meta(self):
        """ Return a dict with all the meta of the file
        """
        document = Poppler.Document.new_from_file(self.uri, self.password)
        metadata = {}
        for key in self.meta_list:
            if document.get_property(key):
                metadata[key] = document.get_property(key)
        return metadata



""" Manage which fileformat can be processed
"""

import archive
import mutagenstripper
import logging
import mat
import misc
import office
import subprocess

STRIPPERS = {
    'application/x-tar': archive.TarStripper,
    'application/x-bzip2': archive.Bzip2Stripper,
    'application/x-gzip': archive.GzipStripper,
    'application/zip': archive.ZipStripper,
    'application/x-bittorrent': misc.TorrentStripper,
    'application/torrent': misc.TorrentStripper,
    'application/opendocument': office.OpenDocumentStripper,
    'application/officeopenxml': office.OpenXmlStripper,
}

logging.basicConfig(level=mat.LOGGING_LEVEL)

# PDF support
pdfSupport = True
try:
    import gi
    gi.require_version('Poppler', '0.18')
    from gi.repository import Poppler
except ImportError:
    logging.error('Unable to import Poppler: no PDF support')
    pdfSupport = False

try:
    import cairo
except ImportError:
    logging.error('Unable to import python-cairo: no PDF support')
    pdfSupport = False

try:
    import pdfrw
except ImportError:
    logging.error('Unable to import python-pdfrw: no PDF support')
    pdfSupport = False

if pdfSupport:
    STRIPPERS['application/x-pdf'] = office.PdfStripper
    STRIPPERS['application/pdf'] = office.PdfStripper


# audio format support with mutagen-python
try:
    import mutagen
    STRIPPERS['audio/x-flac'] = mutagenstripper.FlacStripper
    STRIPPERS['audio/flac'] = mutagenstripper.FlacStripper
    STRIPPERS['audio/vorbis'] = mutagenstripper.OggStripper
    STRIPPERS['audio/ogg'] = mutagenstripper.OggStripper
    STRIPPERS['audio/mpeg'] = mutagenstripper.MpegAudioStripper
except ImportError:
    logging.error('Unable to import python-mutagen: no audio format support')

# exiftool
try:
    subprocess.check_output(['exiftool', '-ver'])
    import exiftool
    STRIPPERS['image/jpeg'] = exiftool.JpegStripper
    STRIPPERS['image/png'] = exiftool.PngStripper
    STRIPPERS['image/tiff'] = exiftool.TiffStripper
except OSError:
    logging.error('Unable to find exiftool: no images support')



#! /usr/bin/python

""" This file is an extension for the Nautilus
    file manager, to provide a contextual menu to
    clean metadata
"""

import logging
import urllib

try:
    import gettext

    gettext.install("mat")
except:
    logging.warning("Failed to initialise gettext")
    _ = lambda x: x

from gi.repository import Nautilus, GObject, Gtk

import libmat.mat
import libmat.strippers


class MatExtension(GObject.GObject, Nautilus.MenuProvider):
    """ A nautilus extension, acting as a frontend to MAT, to clean metadata."""

    def __init__(self):
        logging.debug("nautilus-mat: initialising")

    def get_file_items(self, window, files):
        if len(files) != 1:  # no multi-files support
            return

        current_file = files.pop()

        # We're only going to put ourselves on supported mimetypes' context menus
        if not (current_file.get_mime_type()
                in [i["mimetype"] for i in libmat.mat.list_supported_formats()]):
            logging.debug("%s is not supported by MAT", current_file.get_mime_type())
            return

        # MAT can only handle local file:
        if current_file.get_uri_scheme() != 'file':
            logging.debug("%s files not supported by MAT", current_file.get_uri_scheme())
            return

        # MAT can not clean non-writable files
        if not current_file.can_write():
            logging.debug("%s is not writable by MAT", current_file.get_uri_scheme())
            return

        item = Nautilus.MenuItem(name="Nautilus::clean_metadata",
                                 label=_("Clean metadata"),
                                 tip=_("Clean file's metadata with MAT"),
                                 icon="gtk-clear")
        item.connect('activate', self.menu_activate_cb, current_file)
        return item,

    @staticmethod
    def show_message(message, msg_type=Gtk.MessageType.INFO):
        """ Helper function to show a message in a popup
        :param msg_type: Type of the message to display
        :param message: Message to display
        """
        dialog = Gtk.MessageDialog(parent=None,
                                   flags=Gtk.DialogFlags.MODAL,
                                   type=msg_type,
                                   buttons=Gtk.ButtonsType.OK,
                                   message_format=message)
        ret = dialog.run()
        dialog.destroy()
        return ret

    def menu_activate_cb(self, menu, current_file):
        """ Callback function, used to clean the file
        :param current_file: Name of the selected file
        :param menu: Menu id from which the callback was activated. Unused.
        """
        if file.is_gone():
            return

        # files url in nautilus are starting with 'file://', of length 7
        file_path = urllib.unquote(current_file.get_uri()[7:])

        class_file = libmat.mat.create_class_file(file_path,
                                                  backup=True,
                                                  add2archive=False)
        if class_file:
            if class_file.is_clean():
                self.show_message(_("%s is already clean") % file_path)
            elif not class_file.remove_all():
                self.show_message(_("Unable to clean %s") % file_path, Gtk.MessageType.ERROR)
        else:
            self.show_message(_("Unable to process %s") % file_path, Gtk.MessageType.ERROR)



""" Take care of archives formats
"""

import logging
import os
import shutil
import stat
import tarfile
import tempfile
import zipfile

import mat
import parser

# Zip files do not support dates older than 01/01/1980
ZIP_EPOCH = (1980, 1, 1, 0, 0, 0)


class GenericArchiveStripper(parser.GenericParser):
    """ Represent a generic archive
    """

    def __init__(self, filename, mime, backup, is_writable, **kwargs):
        super(GenericArchiveStripper, self).__init__(filename, mime, backup, is_writable, **kwargs)
        self.compression = ''
        self.add2archive = kwargs['add2archive']
        self.tempdir = tempfile.mkdtemp()

    def __del__(self):
        """ Remove the files inside the temp dir,
            then remove the temp dir
        """
        for root, _, files in os.walk(self.tempdir):
            for item in files:
                path_file = os.path.join(root, item)
                mat.secure_remove(path_file)
        shutil.rmtree(self.tempdir)

    def is_clean(self, list_unsupported=False):
        """ Virtual method to check for harmul metadata
        :param bool list_unsupported:
        """
        raise NotImplementedError

    def list_unsupported(self):
        """ Get a list of every non-supported files present in the archive
        """
        return self.is_clean(list_unsupported=True)

    def remove_all(self):
        """ Virtual method to remove all compromising fields
        """
        raise NotImplementedError


class ZipStripper(GenericArchiveStripper):
    """ Represent a zip file
    """

    @staticmethod
    def __is_zipfile_clean(fileinfo):
        """ Check if a ZipInfo object is clean of metadata added
            by zip itself, independently of the corresponding file metadata
        """
        if fileinfo.comment != '':
            return False
        elif fileinfo.date_time != ZIP_EPOCH:
            return False
        elif fileinfo.create_system != 3:  # 3 is UNIX
            return False
        return True

    def is_clean(self, list_unsupported=False):
        """ Check if the given file is clean from harmful metadata
            When list_unsupported is True, the method returns a list
            of all non-supported/archives files contained in the
            archive.

            :param bool list_unsupported: Should the list of unsupported files be returned
        """
        ret_list = []
        zipin = zipfile.ZipFile(self.filename, 'r')
        if zipin.comment != '' and not list_unsupported:
            logging.debug('%s has a comment', self.filename)
            return False
        for item in zipin.infolist():
            zipin.extract(item, self.tempdir)
            path = os.path.join(self.tempdir, item.filename)
            if not self.__is_zipfile_clean(item) and not list_unsupported:
                logging.debug('%s from %s has compromising zipinfo', item.filename, self.filename)
                return False
            if os.path.isfile(path):
                cfile = mat.create_class_file(path, False, add2archive=self.add2archive)
                if cfile is not None:
                    if not cfile.is_clean():
                        logging.debug('%s from %s has metadata', item.filename, self.filename)
                        if not list_unsupported:
                            return False
                else:
                    logging.info('%s\'s fileformat is not supported or harmless.', item.filename)
                    _, ext = os.path.splitext(path)
                    if os.path.basename(item.filename) not in ('mimetype', '.rels'):
                        if ext not in parser.NOMETA:
                            if not list_unsupported:
                                return False
                            ret_list.append(item.filename)
        zipin.close()
        if list_unsupported:
            return ret_list
        return True

    def get_meta(self):
        """ Return all the metadata of a zip archive"""
        zipin = zipfile.ZipFile(self.filename, 'r')
        metadata = {}
        if zipin.comment != '':
            metadata['comment'] = zipin.comment
        for item in zipin.infolist():
            zipinfo_meta = self.__get_zipinfo_meta(item)
            if zipinfo_meta != {}:  # zipinfo metadata
                metadata[item.filename + "'s zipinfo"] = str(zipinfo_meta)
            zipin.extract(item, self.tempdir)
            path = os.path.join(self.tempdir, item.filename)
            if os.path.isfile(path):
                cfile = mat.create_class_file(path, False, add2archive=self.add2archive)
                if cfile is not None:
                    cfile_meta = cfile.get_meta()
                    if cfile_meta != {}:
                        metadata[item.filename] = str(cfile_meta)
                else:
                    logging.info('%s\'s fileformat is not supported or harmless', item.filename)
        zipin.close()
        return metadata

    @staticmethod
    def __get_zipinfo_meta(zipinfo):
        """ Return all the metadata of a ZipInfo
        """
        metadata = {}
        if zipinfo.comment != '':
            metadata['comment'] = zipinfo.comment
        if zipinfo.date_time != ZIP_EPOCH:
            metadata['modified'] = zipinfo.date_time
        if zipinfo.create_system != 3:  # 3 is UNIX
            metadata['system'] = "windows" if zipinfo.create_system == 2 else "unknown"
        return metadata

    def remove_all(self, whitelist=None, beginning_blacklist=None, ending_blacklist=None):
        """ Remove all metadata from a zip archive, even thoses
            added by Python's zipfile itself. It will not add
            files starting with "begining_blacklist", or ending with
            "ending_blacklist". This method also add files present in
            whitelist to the archive.

            :param list whitelist: Add those files to the produced archive, regardless if they are harmful or not
            :param list beginning_blacklist: If the file starts with $ending_blacklist, it will _not_ be added
            :param list ending_blacklist: If the file end with $ending_blacklist, it will _not_ be added
        """
        if not ending_blacklist:
            ending_blacklist = []
        if not beginning_blacklist:
            beginning_blacklist = []
        if not whitelist:
            whitelist = []
        zipin = zipfile.ZipFile(self.filename, 'r')
        zipout = zipfile.ZipFile(self.output, 'w', allowZip64=True)
        for item in zipin.infolist():
            zipin.extract(item, self.tempdir)
            path = os.path.join(self.tempdir, item.filename)

            beginning = any((True for f in beginning_blacklist if item.filename.startswith(f)))
            ending = any((True for f in ending_blacklist if item.filename.endswith(f)))

            if os.path.isfile(path) and not beginning and not ending:
                cfile = mat.create_class_file(path, False, add2archive=self.add2archive)
                if cfile is not None:
                    # Handle read-only files inside archive
                    old_stat = os.stat(path).st_mode
                    os.chmod(path, old_stat | stat.S_IWUSR)
                    cfile.remove_all()
                    os.chmod(path, old_stat)
                    logging.debug('Processing %s from %s', item.filename, self.filename)
                elif item.filename not in whitelist:
                    logging.info("%s's format is not supported or harmless", item.filename)
                    _, ext = os.path.splitext(path)
                    if not (self.add2archive or ext in parser.NOMETA):
                        continue
                zinfo = zipfile.ZipInfo(item.filename, date_time=ZIP_EPOCH)
                zinfo.compress_type = zipfile.ZIP_DEFLATED
                zinfo.create_system = 3  # Linux
                zinfo.comment = ''
                with open(path, 'r') as f:
                    zipout.writestr(zinfo, f.read())
                # os.utime(path, (ZIP_EPOCH_SECONDS, ZIP_EPOCH_SECONDS))
                # zipout.write(path, item.filename)
        zipin.close()
        zipout.close()

        logging.info('%s processed', self.filename)
        self.do_backup()
        return True


class TarStripper(GenericArchiveStripper):
    """ Represent a tarfile archive
    """

    @staticmethod
    def _remove_tar_added(current_file):
        """ Remove the meta added by tarfile itself to the file
        """
        current_file.mtime = 0
        current_file.uid = 0
        current_file.gid = 0
        current_file.uname = ''
        current_file.gname = ''
        return current_file

    def remove_all(self, whitelist=None):
        """ Remove all harmful metadata from the tarfile.
            The method will also add every files matching
            whitelist in the produced archive.
            :param list whitelist: Files to add the to produced archive,
                    regardless if they are considered harmfull.
        """
        if not whitelist:
            whitelist = []
        tarin = tarfile.open(self.filename, 'r' + self.compression, encoding='utf-8')
        tarout = tarfile.open(self.output, 'w' + self.compression, encoding='utf-8')
        for item in tarin.getmembers():
            tarin.extract(item, self.tempdir)
            if item.isfile():
                path = os.path.join(self.tempdir, item.name)
                cfile = mat.create_class_file(path, False, add2archive=self.add2archive)
                if cfile is not None:
                    # Handle read-only files inside archive
                    old_stat = os.stat(path).st_mode
                    os.chmod(path, old_stat | stat.S_IWUSR)
                    cfile.remove_all()
                    os.chmod(path, old_stat)
                elif self.add2archive or os.path.splitext(item.name)[1] in parser.NOMETA:
                    logging.debug("%s' format is either not supported or harmless", item.name)
                elif item.name in whitelist:
                    logging.debug('%s is not supported, but MAT was told to add it anyway.', item.name)
                else:  # Don't add the file to the archive
                    logging.debug('%s will not be added', item.name)
                    continue
                tarout.add(unicode(path.decode('utf-8')),
                           unicode(item.name.decode('utf-8')),
                           filter=self._remove_tar_added)
        tarin.close()
        tarout.close()
        self.do_backup()
        return True

    @staticmethod
    def is_file_clean(current_file):
        """ Check metadatas added by tarfile
        :param tarfile.TarInfo current_file:
        """
        if current_file.mtime != 0:
            return False
        elif current_file.uid != 0:
            return False
        elif current_file.gid != 0:
            return False
        elif current_file.uname != '':
            return False
        elif current_file.gname != '':
            return False
        return True

    def is_clean(self, list_unsupported=False):
        """ Check if the file is clean from harmful metadatas
            When list_unsupported is True, the method returns a list
            of all non-supported/archives files contained in the
            archive.
            :param bool list_unsupported:
        """
        ret_list = []
        tarin = tarfile.open(self.filename, 'r' + self.compression)
        for item in tarin.getmembers():
            if not self.is_file_clean(item) and not list_unsupported:
                logging.debug('%s from %s has compromising tarinfo', item.name, self.filename)
                return False
            tarin.extract(item, self.tempdir)
            path = os.path.join(self.tempdir, item.name)
            if item.isfile():
                cfile = mat.create_class_file(path, False, add2archive=self.add2archive)
                if cfile is not None:
                    if not cfile.is_clean():
                        logging.debug('%s from %s has metadata', item.name.decode("utf8"), self.filename)
                        if not list_unsupported:
                            return False
                        # Nested archives are treated like unsupported files
                        elif isinstance(cfile, GenericArchiveStripper):
                            ret_list.append(item.name)
                else:
                    logging.info("%s's format is not supported or harmless", item.name)
                    if os.path.splitext(path)[1] not in parser.NOMETA:
                        if not list_unsupported:
                            return False
                        ret_list.append(item.name)
        tarin.close()
        if list_unsupported:
            return ret_list
        return True

    def get_meta(self):
        """ Return a dict with all the meta of the tarfile
        """
        tarin = tarfile.open(self.filename, 'r' + self.compression)
        metadata = {}
        for item in tarin.getmembers():
            current_meta = {}
            if item.isfile():
                tarin.extract(item, self.tempdir)
                path = os.path.join(self.tempdir, item.name)
                class_file = mat.create_class_file(path, False, add2archive=self.add2archive)
                if class_file is not None:
                    meta = class_file.get_meta()
                    if meta:
                        current_meta['file'] = str(meta)
                else:
                    logging.info("%s's format is not supported or harmless", item.name)

                if not self.is_file_clean(item):  # if there is meta
                    current_meta['mtime'] = item.mtime
                    current_meta['uid'] = item.uid
                    current_meta['gid'] = item.gid
                    current_meta['uname'] = item.uname
                    current_meta['gname'] = item.gname
                    metadata[item.name] = str(current_meta)
        tarin.close()
        return metadata


class TerminalZipStripper(ZipStripper):
    """ Represent a terminal level archive.
        This type of archive can not contain nested archives.
        It is used for formats like docx, which are basically
        ziped xml.
    """
    pass


class GzipStripper(TarStripper):
    """ Represent a tar.gz archive
    """

    def __init__(self, filename, mime, backup, is_writable, **kwargs):
        super(GzipStripper, self).__init__(filename, mime, backup, is_writable, **kwargs)
        self.compression = ':gz'


class Bzip2Stripper(TarStripper):
    """ Represent a tar.bz2 archive
    """

    def __init__(self, filename, mime, backup, is_writable, **kwargs):
        super(Bzip2Stripper, self).__init__(filename, mime, backup, is_writable, **kwargs)
        self.compression = ':bz2'



""" Base exceptions for MAT
"""


class UnableToRemoveFile(Exception):
    """This exception is raised when a file could not be removed
    """
    pass


class UnableToWriteFile(Exception):
    """This exception is raised when a file
        can could not be chmod +w
    """
    pass



""" Parent class of all parser
"""

import os
import shutil
import tempfile


import mat

NOMETA = frozenset((
    '.bmp',   # "raw" image
    '.rdf',   # text
    '.txt',   # plain text
    '.xml',   # formated text (XML)
    '.rels',  # openXML formated text
))

FIELD = object()


class GenericParser(object):
    """ Parent class of all parsers
    """
    def __init__(self, filename, mime, backup, is_writable, **kwargs):
        self.filename = ''
        self.mime = mime
        self.backup = backup
        self.is_writable = is_writable
        self.filename = filename
        self.basename = os.path.basename(filename)
        self.output = tempfile.mkstemp()[1]

    def __del__(self):
        """ Remove tempfile if it was not used
        """
        if os.path.exists(self.output):
            mat.secure_remove(self.output)

    def is_clean(self):
        """
            Check if the file is clean from harmful metadatas
        """
        raise NotImplementedError

    def remove_all(self):
        """ Remove all compromising fields
        """
        raise NotImplementedError

    def create_backup_copy(self):
        """ Create a backup copy
        """
        shutil.copy2(self.filename, os.path.join(self.filename, '.bak'))

    def do_backup(self):
        """ Keep a backup of the file if asked.

            The process of double-renaming is not very elegant,
            but it greatly simplify new strippers implementation.
        """
        if self.backup:
            shutil.move(self.filename, os.path.join(self.filename, '.bak'))
        else:
            mat.secure_remove(self.filename)
        shutil.move(self.output, self.filename)



#!/usr/bin/env python

""" Metadata anonymisation toolkit library
"""

import logging
import mimetypes
import os
import platform
import subprocess
import xml.sax

import libmat.exceptions

__version__ = '0.6.1'
__author__ = 'jvoisin'

# Silence
LOGGING_LEVEL = logging.ERROR
logging.basicConfig(filename='', level=LOGGING_LEVEL)

import strippers  # this is loaded here because we need LOGGING_LEVEL


def get_logo():  # pragma: no cover
    """ Return the path to the logo
    """
    if os.path.isfile(os.path.join(os.path.curdir, 'data/mat.png')):
        return os.path.join(os.path.curdir, 'data/mat.png')
    elif os.path.isfile('/usr/share/pixmaps/mat.png'):
        return '/usr/share/pixmaps/mat.png'
    elif os.path.isfile('/usr/local/share/pixmaps/mat.png'):
        return '/usr/local/share/pixmaps/mat.png'


def get_datafile_path(filename):  # pragma: no cover
    """ Return the path to $filename
    :param string filename:
    """
    paths = ['data', '/usr/local/share/mat/', '/usr/share/mat/']
    for path in paths:
        filepath = os.path.join(os.path.curdir, path, filename)
        if os.path.isfile(filepath):
            return filepath


def list_supported_formats():  # pragma: no cover
    """ Return a list of all locally supported fileformat.
        It parses that FORMATS file, and removes locally
        non-supported formats.
    """
    handler = XMLParser()
    parser = xml.sax.make_parser()
    parser.setContentHandler(handler)
    path = get_datafile_path('FORMATS')
    with open(path, 'r') as xmlfile:
        parser.parse(xmlfile)

    localy_supported = []
    for item in handler.list:
        if item['mimetype'].split(',')[0] in strippers.STRIPPERS:
            localy_supported.append(item)

    return localy_supported


class XMLParser(xml.sax.handler.ContentHandler):  # pragma: no cover
    """ Parse the supported format xml, and return a corresponding
        list of dict
    """

    def __init__(self):
        xml.sax.handler.ContentHandler.__init__(self)
        self.dict = {}
        self.list = []
        self.content, self.key = '', ''
        self.between = False

    def startElement(self, name, attrs):
        """ Called when entering into xml tag
        :param attrs: Attributes of the `nam` xml tag
        :param name: Name of the xml tag
        """
        self.between = True
        self.key = name
        self.content = ''

    def endElement(self, name):
        """ Called when exiting a xml tag
        :param name: name of the element
        """
        if name == 'format':  # leaving a fileformat section
            self.list.append(self.dict.copy())
            self.dict.clear()
        else:
            content = self.content.replace('\s', ' ')
            self.dict[self.key] = content
            self.between = False

    def characters(self, characters):
        """ Concatenate the content between opening and closing tags
        :param characters: content of the tag
        """
        if self.between:
            self.content += characters


def secure_remove(filename):
    """ Securely remove $filename
    :param str filename: File to be removed
    """
    try:  # I want the file removed, even if it's read-only
        os.chmod(filename, 220)
    except OSError:
        logging.error('Unable to add write rights to %s', filename)
        raise libmat.exceptions.UnableToWriteFile

    try:
        shred = 'shred'
        if platform.system() == 'MacOS':
            shred = 'gshred'
        if not subprocess.call([shred, '--remove', filename]):
            return True
        else:
            raise OSError
    except OSError:
        logging.error('Unable to securely remove %s', filename)

    try:
        os.remove(filename)
    except OSError:
        logging.error('Unable to remove %s', filename)
        raise libmat.exceptions.UnableToRemoveFile

    return True


def create_class_file(name, backup, **kwargs):
    """ Return a $FILETYPEStripper() class,
        corresponding to the filetype of the given file

        :param str name: name of the file to be parsed
        :param bool backup: shell the file be backuped?
    """
    if not os.path.isfile(name):  # check if the file exists
        logging.error('%s is not a valid file', name)
        return None
    elif not os.access(name, os.R_OK):  # check read permissions
        logging.error('%s is is not readable', name)
        return None

    mime = mimetypes.guess_type(name)[0]
    if not mime:
        logging.info('Unable to find mimetype of %s', name)
        return None

    if mime.startswith('application/vnd.oasis.opendocument'):
        mime = 'application/opendocument'  # opendocument fileformat
    elif mime.startswith('application/vnd.openxmlformats-officedocument'):
        mime = 'application/officeopenxml'  # office openxml

    is_writable = os.access(name, os.W_OK)

    try:
        stripper_class = strippers.STRIPPERS[mime]
    except KeyError:
        logging.info('Don\'t have stripper for %s format', mime)
        return None

    return stripper_class(name, mime, backup, is_writable, **kwargs)



# Copyright 2007 by Petru Paler
# Copyright 2011 by Julien (jvoisin) Voisin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""
    A quick (and also nice) lib to bencode/bdecode torrent files
"""


class BTFailure(Exception):
    """Custom Exception"""
    pass


class Bencached(object):
    """Custom type : cached string"""
    __slots__ = ['bencoded']

    def __init__(self, string):
        self.bencoded = string


def decode_int(x, f):
    """decode an int"""
    f += 1
    newf = x.index('e', f)
    if x[f:f + 1] == '-0':
        raise ValueError
    elif x[f] == '0' and newf != f + 1:
        raise ValueError
    return int(x[f:newf]), newf + 1


def decode_string(x, f):
    """decode a string"""
    colon = x.index(':', f)
    if x[f] == '0' and colon != f + 1:
        raise ValueError
    n = int(x[f:colon])
    colon += 1
    return x[colon:colon + n], colon + n


def decode_list(x, f):
    """decode a list"""
    result = []
    f += 1
    while x[f] != 'e':
        v, f = DECODE_FUNC[x[f]](x, f)
        result.append(v)
    return result, f + 1


def decode_dict(x, f):
    """decode a dict"""
    result = {}
    f += 1
    while x[f] != 'e':
        k, f = decode_string(x, f)
        result[k], f = DECODE_FUNC[x[f]](x, f)
    return result, f + 1


def encode_bool(x, r):
    """bencode a boolean"""
    encode_int(1 if r else 0, r)


def encode_int(x, r):
    """bencode an integer/float"""
    r.extend(('i', str(x), 'e'))


def encode_list(x, r):
    """bencode a list/tuple"""
    r.append('l')
    [ENCODE_FUNC[type(item)](item, r) for item in x]
    r.append('e')


def encode_dict(x, result):
    """bencode a dict"""
    result.append('d')
    ilist = list(x.items())
    ilist.sort()
    for k, v in ilist:
        result.extend((str(len(k)), ':', k))
        ENCODE_FUNC[type(v)](v, result)
    result.append('e')


DECODE_FUNC = {str(x): decode_string for x in range(9)}
DECODE_FUNC['l'] = decode_list
DECODE_FUNC['d'] = decode_dict
DECODE_FUNC['i'] = decode_int

ENCODE_FUNC = {}
ENCODE_FUNC[Bencached] = lambda x, r: r.append(x.bencoded)
ENCODE_FUNC[int] = encode_int
ENCODE_FUNC[int] = encode_int
ENCODE_FUNC[bytes] = lambda x, r: r.extend((str(len(x)), ':', x))
ENCODE_FUNC[list] = encode_list
ENCODE_FUNC[tuple] = encode_list
ENCODE_FUNC[dict] = encode_dict
ENCODE_FUNC[bool] = encode_bool


def bencode(string):
    """bencode $string"""
    table = []
    ENCODE_FUNC[type(string)](string, table)
    return ''.join(table)


def bdecode(string):
    """decode $string"""
    try:
        result, lenght = DECODE_FUNC[string[0]](string, 0)
    except (IndexError, KeyError, ValueError):
        raise BTFailure('Not a valid bencoded string')
    if lenght != len(string):
        raise BTFailure('Invalid bencoded value (data after valid prefix)')
    return result



""" Take care of mutagen-supported formats (audio)
"""

import parser

from mutagen.flac import FLAC
from mutagen.oggvorbis import OggVorbis
from mutagen.mp3 import MP3


class MutagenStripper(parser.GenericParser):
    """ Parser using the (awesome) mutagen library. """
    def __init__(self, filename, mime, backup, is_writable, **kwargs):
        super(MutagenStripper, self).__init__(filename, mime, backup, is_writable, **kwargs)
        self.mfile = None  # This will be instanciated in self._create_mfile()
        self._create_mfile()

    def _create_mfile(self):
        """ This method must be overrriden to instanciate the `mfile` attribute."""
        raise NotImplementedError

    def is_clean(self):
        """ Check if the file is clean. """
        return not self.mfile.tags

    def remove_all(self):
        """ Remove all harmful metadata. """
        if self.backup:
            self.create_backup_copy()
        self.mfile.delete()
        try:  # mutagen will choke on files without a metadata block
            self.mfile.save()
        except ValueError:
            pass
        return True

    def get_meta(self):
        """
            Return the content of the metadata block is present
        """
        metadata = {}
        if self.mfile.tags:
            for key, value in self.mfile.tags:
                metadata[key] = value
        return metadata


class MpegAudioStripper(MutagenStripper):
    """ Represent a mp3 vorbis file
    """
    def _create_mfile(self):
        self.mfile = MP3(self.filename)

    def get_meta(self):
        """
            Return the content of the metadata block is present
        """
        metadata = {}
        if self.mfile.tags:
            for key in self.mfile.tags.keys():
                meta = self.mfile.tags[key]
                try:  # Sometimes, the field has a human-redable description
                    desc = meta.desc
                except AttributeError:
                    desc = key
                text = meta.text[0]
                metadata[desc] = text
        return metadata


class OggStripper(MutagenStripper):
    """ Represent an ogg vorbis file
    """
    def _create_mfile(self):
        self.mfile = OggVorbis(self.filename)


class FlacStripper(MutagenStripper):
    """ Represent a Flac audio file
    """
    def _create_mfile(self):
        self.mfile = FLAC(self.filename)

    def remove_all(self):
        """ Remove the "metadata" block from the file
        """
        super(FlacStripper, self).remove_all()
        self.mfile.clear_pictures()
        self.mfile.save()
        return True

    def is_clean(self):
        """ Check if the "metadata" block is present in the file
        """
        return super(FlacStripper, self).is_clean() and not self.mfile.pictures

    def get_meta(self):
        """ Return the content of the metadata block if present
        """
        metadata = super(FlacStripper, self).get_meta()
        if self.mfile.pictures:
            metadata['picture:'] = 'yes'
        return metadata



#!/usr/bin/env python
# -*- coding: utf-8 -*

"""
    Unit test for the CLI interface
"""

import os
import unittest
import subprocess
import sys
import tarfile
import stat

import test
from libmat import mat


class TestRemovecli(test.MATTest):
    """
        test if cli correctly remove metadatas
    """

    def test_remove(self):
        """make sure that the cli remove all compromizing meta"""
        for _, dirty in self.file_list:
            subprocess.call(['mat', '--add2archive', dirty])
            current_file = mat.create_class_file(dirty, False, add2archive=True, low_pdf_quality=True)
            self.assertTrue(current_file.is_clean())

    def test_remove_fileformat_specific_options(self):
        """ test metadata removal with fileformat-specific options """
        for _, dirty in self.file_list:  # can't be faster than that :/
            if dirty.endswith('pdf'):
                subprocess.call(['mat', '--low-pdf-quality', dirty])
                current_file = mat.create_class_file(dirty, False, low_pdf_quality=True)
                self.assertTrue(current_file.is_clean())

    def test_remove_empty(self):
        """Test removal with clean files\n"""
        for clean, _ in self.file_list:
            subprocess.call(['mat', '--add2archive', clean])
            current_file = mat.create_class_file(clean, False, add2archive=True, low_pdf_quality=True)
            self.assertTrue(current_file.is_clean())


class TestListcli(test.MATTest):
    """
        test if cli correctly display metadatas
    """

    def test_list_clean(self):
        """check if get_meta returns meta"""
        for clean, _ in self.file_list:
            proc = subprocess.Popen(['mat', '-d', clean],
                                    stdout=subprocess.PIPE)
            stdout, _ = proc.communicate()
            self.assertEqual(str(stdout).strip('\n'), "[+] File %s \
:\nNo harmful metadata found" % clean)

    def test_list_dirty(self):
        """check if get_meta returns all the expected meta"""
        for _, dirty in self.file_list:
            proc = subprocess.Popen(['mat', '-d', dirty],
                                    stdout=subprocess.PIPE)
            stdout, _ = proc.communicate()
            self.assertNotEqual(str(stdout), "[+] File %s :\n No\
harmul metadata found" % dirty)


class TestisCleancli(test.MATTest):
    """
        check if cli correctly check if a file is clean or not
    """

    def test_clean(self):
        """test is_clean on clean files"""
        for clean, _ in self.file_list:
            proc = subprocess.Popen(['mat', '-c', clean],
                                    stdout=subprocess.PIPE)
            stdout, _ = proc.communicate()
            self.assertEqual(str(stdout).strip('\n'), '[+] %s is clean' % clean)

    def test_dirty(self):
        """test is_clean on dirty files"""
        for _, dirty in self.file_list:
            proc = subprocess.Popen(['mat', '-c', dirty],
                                    stdout=subprocess.PIPE)
            stdout, _ = proc.communicate()
            self.assertEqual(str(stdout).strip('\n'), '[+] %s is not clean' % dirty)


class TestFileAttributes(unittest.TestCase):
    """
        test various stuffs about files (readable, writable, exist, ...)
    """

    def test_not_writtable(self):
        """ test MAT's behaviour on non-writable file"""
        proc = subprocess.Popen(['mat', 'not_writtable'],
                                stdout=subprocess.PIPE)
        stdout, _ = proc.communicate()
        self.assertEqual(str(stdout).strip('\n'), '[-] Unable to process not_writtable')

    def test_not_exist(self):
        """ test MAT's behaviour on non-existent file"""
        proc = subprocess.Popen(['mat', 'ilikecookies'],
                                stdout=subprocess.PIPE)
        stdout, _ = proc.communicate()
        self.assertEqual(str(stdout).strip('\n'), '[-] Unable to process ilikecookies')

    def test_empty(self):
        """ test MAT's behaviour on empty file"""
        proc = subprocess.Popen(['mat', 'empty_file'], stdout=subprocess.PIPE)
        stdout, _ = proc.communicate()
        self.assertEqual(str(stdout).strip('\n'), '[-] Unable to process empty_file')

    def test_not_readable(self):
        """ test MAT's behaviour on non-writable file"""
        open('non_readable', 'a').close()
        os.chmod('non_readable', 0 & stat.S_IWRITE)
        proc = subprocess.Popen(['mat', 'non_readable'], stdout=subprocess.PIPE)
        stdout, _ = proc.communicate()
        os.remove('non_readable')


class TestUnsupported(test.MATTest):
    """ test MAT's behaviour on unsupported files """
    def test_abort_unsupported(self):
        """ test if the cli aborts on unsupported files
        """
        tarpath = os.path.join(self.tmpdir, "test.tar.bz2")
        tar = tarfile.open(tarpath, "w")
        for f in ('libtest.py', 'test.py', 'clitest.py'):
            tar.add(f, f)
        tar.close()
        proc = subprocess.Popen(['mat', tarpath], stdout=subprocess.PIPE)
        stdout, _ = proc.communicate()
        self.assertTrue('It contains unsupported filetypes:'
                        '\n- libtest.py\n- test.py\n- clitest.py\n'
                        in str(stdout))


class TestHelp(test.MATTest):
    """ Test the different ways to trigger help """
    def test_dash_h(self):
        """ test help invocation with `-h` and `--help` """
        proc = subprocess.Popen(['mat', '-h'], stdout=subprocess.PIPE)
        stdout, _ = proc.communicate()
        self.assertTrue('show this help message and exit' in stdout)

        proc = subprocess.Popen(['mat', '--help'], stdout=subprocess.PIPE)
        stdout, _ = proc.communicate()
        self.assertTrue('show this help message and exit' in stdout)

    def test_no_argument(self):
        """ test help invocation when no argument is provided """
        proc = subprocess.Popen(['mat'], stdout=subprocess.PIPE)
        stdout, _ = proc.communicate()
        self.assertTrue('show this help message and exit' in stdout)

    def test_wrong_argument(self):
        """ Test MAT's behaviour on wrong argument """
        proc = subprocess.Popen(['mat', '--obviously-wrong-argument'], stderr=subprocess.PIPE)
        _, stderr = proc.communicate()
        self.assertTrue(('usage: mat [-h]' and ' error: unrecognized arguments:') in stderr)


def get_tests():
    """ Return every clitests"""
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestRemovecli))
    suite.addTest(unittest.makeSuite(TestListcli))
    suite.addTest(unittest.makeSuite(TestisCleancli))
    suite.addTest(unittest.makeSuite(TestUnsupported))
    suite.addTest(unittest.makeSuite(TestFileAttributes))
    suite.addTest(unittest.makeSuite(TestHelp))
    return suite



""" Care about images with help of the amazing (perl) library Exiftool.
"""

import subprocess
import parser


class ExiftoolStripper(parser.GenericParser):
    """ A generic stripper class using exiftool as backend
    """

    def __init__(self, filename, mime, backup, is_writable, **kwargs):
        super(ExiftoolStripper, self).__init__(filename, mime, backup, is_writable, **kwargs)
        self.allowed = {'ExifTool Version Number', 'File Name', 'Directory', 'File Size', 'File Modification Date/Time',
                        'File Access Date/Time', 'File Permissions', 'File Type', 'File Type Extension', 'MIME Type',
                        'Image Width', 'Image Height', 'Image Size', 'File Inode Change Date/Time', 'Megapixels'}
        self._set_allowed()

    def _set_allowed(self):
        """ Virtual method. Set the allowed/harmless list of metadata
        """
        raise NotImplementedError

    def remove_all(self):
        """ Remove all metadata with help of exiftool
        """
        try:
            if self.backup:
                self.create_backup_copy()
            # Note: '-All=' must be followed by a known exiftool option.
            # Also, '-CommonIFD0' is needed for .tiff files
            subprocess.call(['exiftool', '-all=', '-adobe=', '-exif:all=', '-Time:All=', '-m',
                             '-CommonIFD0=', '-overwrite_original', self.filename],
                            stdout=open('/dev/null'), stderr=open('/dev/null'))
            return True
        except OSError:
            return False

    def is_clean(self):
        """ Check if the file is clean with the help of exiftool
        """
        return not self.get_meta()

    def get_meta(self):
        """ Return every harmful meta with help of exiftool.
            Exiftool output looks like this:
            field name : value
            field name : value
        """
        output = subprocess.Popen(['exiftool', self.filename],
                                  stdout=subprocess.PIPE).communicate()[0]
        meta = {}
        for i in output.split('\n')[:-1]:  # chop last char ('\n')
            key = i.split(':')[0].strip()
            if key not in self.allowed:
                meta[key] = i.split(':')[1].strip()  # add the field name to the metadata set
        return meta


class JpegStripper(ExiftoolStripper):
    """ Care about jpeg files with help
        of exiftool
    """

    def _set_allowed(self):
        self.allowed.update(['JFIF Version', 'Resolution Unit',
                             'X Resolution', 'Y Resolution', 'Encoding Process',
                             'Bits Per Sample', 'Color Components', 'Y Cb Cr Sub Sampling'])


class PngStripper(ExiftoolStripper):
    """ Care about png files with help
        of exiftool
    """

    def _set_allowed(self):
        self.allowed.update(['Bit Depth', 'Color Type',
                             'Compression', 'Filter', 'Interlace', 'Palette',
                             'Pixels Per Unit X',
                             'Pixels Per Unit Y', 'Pixel Units', 'Significant Bits',
                             'Background Color', 'SRGB Rendering'])


class TiffStripper(ExiftoolStripper):
    """ Care about tiff files with help
        of exiftool
    """

    def _set_allowed(self):
        # Todo: it would be awesome to detect the Resolution Unit, and to transform it in centimeter if it's in inches.
        self.allowed.update(['X Resolution', 'Y Resolution', 'Compression', 'Bits Per Sample',
                             'Strip Offsets', 'Photometric Interpretation', 'Strip Byte Counts',
                             'Resolution Unit', 'Exif Byte Order', 'Samples Per Pixel', 'Rows Per Strip',
                             'Orientation'])



#!/usr/bin/env python
# -*- coding: utf-8 -*

"""
    Unit test for the library
"""

import os
import sys
import stat
import shutil
import tarfile
import tempfile
import unittest

import test
import libmat


class TestRemovelib(test.MATTest):
    """ test the remove_all() method
    """

    def test_remove(self):
        """make sure that the lib remove all compromizing meta"""
        for _, dirty in self.file_list:
            current_file = libmat.mat.create_class_file(dirty, False, add2archive=True)
            current_file.remove_all()
            current_file = libmat.mat.create_class_file(dirty, False, add2archive=True)
            self.assertTrue(current_file.is_clean())

    def test_remove_fileformat_specific_options(self):
        """ test metadata removal with fileformat-specific options """
        for _, dirty in self.file_list:  # can't be faster than that :/
            if dirty.endswith('pdf'):
                current_file = libmat.mat.create_class_file(dirty, False, add2archive=True, low_pdf_quality=True)
                current_file.remove_all()
                current_file = libmat.mat.create_class_file(dirty, False, add2archive=True, low_pdf_quality=True)
                self.assertTrue(current_file.is_clean())

    def test_remove_empty(self):
        """Test removal with clean files"""
        for clean, _ in self.file_list:
            current_file = libmat.mat.create_class_file(clean, False, add2archive=True)
            current_file.remove_all()
            current_file = libmat.mat.create_class_file(clean, False, add2archive=True)
            self.assertTrue(current_file.is_clean())


class TestListlib(test.MATTest):
    """ test the get_meta() method
    """

    def test_list(self):
        """check if get_meta returns metadata"""
        for _, dirty in self.file_list:
            current_file = libmat.mat.create_class_file(dirty, False, add2archive=True)
            self.assertIsNotNone(current_file.get_meta())

    def testlist_list_empty(self):
        """check that a listing of a clean file returns an empty dict"""
        for clean, _ in self.file_list:
            current_file = libmat.mat.create_class_file(clean, False, add2archive=True)
            self.assertEqual(current_file.get_meta(), dict())


class TestisCleanlib(test.MATTest):
    """ Test the is_clean() method
    """

    def test_dirty(self):
        """test is_clean on dirty files"""
        for _, dirty in self.file_list:
            current_file = libmat.mat.create_class_file(dirty, False, add2archive=True)
            self.assertFalse(current_file.is_clean())

    def test_clean(self):
        """test is_clean on clean files"""
        for clean, _ in self.file_list:
            current_file = libmat.mat.create_class_file(clean, False, add2archive=True)
            self.assertTrue(current_file.is_clean())


class TestFileAttributes(unittest.TestCase):
    """
        test various stuffs about files (readable, writable, exist, ...)
    """

    def test_not_exist(self):
        """ test MAT's behaviour on non-existent file"""
        self.assertFalse(libmat.mat.create_class_file('non_existent_file', False, add2archive=True))

    def test_empty(self):
        """ test MAT's behaviour on empty file"""
        open('empty_file', 'a').close()
        self.assertFalse(libmat.mat.create_class_file('empty_file', False, add2archive=True))
        os.remove('empty_file')

    def test_not_writtable(self):
        """ test MAT's behaviour on non-writable file"""
        self.assertFalse(libmat.mat.create_class_file('not_writtable', False, add2archive=True))

    def test_not_readable(self):
        """ test MAT's behaviour on non-readable file"""
        open('non_readable', 'a').close()
        os.chmod('non_readable', 0 | stat.S_IWRITE)
        self.assertFalse(libmat.mat.create_class_file('non_readable', False, add2archive=True))
        os.remove('non_readable')


class TestSecureRemove(unittest.TestCase):
    """ Test the secure_remove function
    """

    def test_remove_existing(self):
        """ test the secure removal of an existing file
        """
        _, file_to_remove = tempfile.mkstemp()
        self.assertTrue(libmat.mat.secure_remove(file_to_remove))

    def test_remove_fail(self):
        """ test the secure removal of an non-removable file
        """
        self.assertRaises(libmat.exceptions.UnableToWriteFile, libmat.mat.secure_remove, '/NOTREMOVABLE')


class TestArchiveProcessing(test.MATTest):
    """ Test archives processing
    """

    def test_remove_bz2(self):
        """ Test MAT's ability to process .tar.bz2
        """
        tarpath = os.path.join(self.tmpdir, "test.tar.bz2")
        tar = tarfile.open(tarpath, "w:bz2")
        for clean, dirty in self.file_list:
            tar.add(dirty)
            tar.add(clean)
        tar.close()
        current_file = libmat.mat.create_class_file(tarpath, False, add2archive=False)
        current_file.remove_all()
        current_file = libmat.mat.create_class_file(tarpath, False, add2archive=False)
        self.assertTrue(current_file.is_clean())

    def test_remove_tar(self):
        """ Test MAT on tar files
        """
        tarpath = os.path.join(self.tmpdir, "test.tar")
        tar = tarfile.open(tarpath, "w")
        for clean, dirty in self.file_list:
            tar.add(dirty)
            tar.add(clean)
        tar.close()
        current_file = libmat.mat.create_class_file(tarpath, False, add2archive=False)
        current_file.remove_all()
        current_file = libmat.mat.create_class_file(tarpath, False, add2archive=False)
        self.assertTrue(current_file.is_clean())

    def test_remove_gz(self):
        """ Test MAT on tar.gz files
        """
        tarpath = os.path.join(self.tmpdir, "test.tar.gz")
        tar = tarfile.open(tarpath, "w")
        for clean, dirty in self.file_list:
            tar.add(dirty)
            tar.add(clean)
        tar.close()
        current_file = libmat.mat.create_class_file(tarpath, False, add2archive=False)
        current_file.remove_all()
        current_file = libmat.mat.create_class_file(tarpath, False, add2archive=False)
        self.assertTrue(current_file.is_clean())

    def test_get_unsupported(self):
        """ Test the get_unsupported feature, used by the GUI
        """
        tarpath = os.path.join(self.tmpdir, "test.tar.bz2")
        tar = tarfile.open(tarpath, "w")
        for f in ('libtest.py', 'test.py', 'clitest.py'):
            tar.add(f, f)
        tar.close()
        current_file = libmat.mat.create_class_file(tarpath, False, add2archive=False)
        unsupported_files = set(current_file.is_clean(list_unsupported=True))
        self.assertEqual(unsupported_files, {'libtest.py', 'test.py', 'clitest.py'})

    def test_archive_unwritable_content(self):
        path = os.path.join(self.tmpdir, './unwritable_content.zip')
        shutil.copy2('./unwritable_content.zip', self.tmpdir)
        current_file = libmat.mat.create_class_file(path, False, add2archive=False)
        current_file.remove_all()
        current_file = libmat.mat.create_class_file(path, False, add2archive=False)
        self.assertTrue(current_file.is_clean())


def get_tests():
    """ Returns every libtests"""
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestRemovelib))
    suite.addTest(unittest.makeSuite(TestListlib))
    suite.addTest(unittest.makeSuite(TestisCleanlib))
    suite.addTest(unittest.makeSuite(TestFileAttributes))
    suite.addTest(unittest.makeSuite(TestSecureRemove))
    suite.addTest(unittest.makeSuite(TestArchiveProcessing))
    return suite



#!/usr/bin/env python

import os

from distutils.core import setup, Command
from DistUtilsExtra.command import *

__version__ = '0.6.1'

# Remove MANIFEST file, since distutils
# doesn't properly update it when
# the contents of directories changes.
if os.path.exists('MANIFEST'):
    os.remove('MANIFEST')


class PyTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.chdir('test')
        import test
        test.test.set_local()
        test.test.run_all_tests()

setup(
    name='MAT',
    version=__version__,
    description='Metadata Anonymisation Toolkit',
    long_description='A Metadata Anonymisation Toolkit in Python',
    author='jvoisin',
    author_email='julien.voisin@dustri.org',
    platforms='linux',
    license='GPLv2',
    url='https://mat.boum.org',
    packages=['libmat', 'libmat.bencode'],
    scripts=['mat', 'mat-gui'],
    data_files=[
        ('share/applications', ['mat.desktop']),
        ('share/mat', ['data/FORMATS', 'data/mat.glade']),
        ('share/pixmaps', ['data/mat.png']),
        ('share/doc/mat', ['README.md', 'README.security']),
        ('share/man/man1', ['mat.1', 'mat-gui.1']),
        ('share/nautilus-python/extensions', ['nautilus/nautilus-mat.py'])
    ],
    cmdclass={
        'test': PyTest,
        'build': build_extra.build_extra,
        'build_i18n': build_i18n.build_i18n,
        'build_help': build_help.build_help,
        'build_icons': build_icons.build_icons,
        'clean': clean_i18n.clean_i18n,
    },
    requires=['mutagen', 'gi', 'pdfrw']
)



""" Care about misc formats
"""

import parser

from bencode import bencode


class TorrentStripper(parser.GenericParser):
    """ Represent a torrent file with the help
        of the bencode lib from Petru Paler
    """

    def __init__(self, filename, mime, backup, is_writable, **kwargs):
        super(TorrentStripper, self).__init__(filename, mime, backup, is_writable, **kwargs)
        self.fields = frozenset(['announce', 'info', 'name', 'path', 'piece length', 'pieces',
                                 'length', 'files', 'announce-list', 'nodes', 'httpseeds', 'private', 'root hash'])

    def __get_key_recursively(self, dictionary):
        """ Get recursively all keys from a dict and
            its subdicts
        """
        for i, j in list(dictionary.items()):
            if isinstance(j, dict):
                return {i}.union(self.__get_key_recursively(j))
            return {i}

    def is_clean(self):
        """ Check if the file is clean from harmful metadata
        """
        with open(self.filename, 'r') as f:
            decoded = bencode.bdecode(f.read())
        return self.fields.issuperset(self.__get_key_recursively(decoded))

    def __get_meta_recursively(self, dictionary):
        """ Get recursively all harmful metadata
        """
        d = dict()
        for i, j in list(dictionary.items()):
            if i not in self.fields:
                d[i] = j
            elif isinstance(j, dict):
                d = dict(d.items() + list(self.__get_meta_recursively(j).items()))
        return d

    def get_meta(self):
        """ Return a dict with all the meta of the file
        """
        with open(self.filename, 'r') as f:
            decoded = bencode.bdecode(f.read())
        return self.__get_meta_recursively(decoded)

    def __remove_all_recursively(self, dictionary):
        """ Remove recursively all compromizing fields
        """
        d = dict()
        for i, j in list(dictionary.items()):
            if i not in self.fields:
                continue
            elif isinstance(j, dict):
                d = dict(list(d.items()) + list(self.__get_meta_recursively(j).items()))
            else:
                d[i] = j
        return d

    def remove_all(self):
        """ Remove all comprimizing fields
        """
        with open(self.filename, 'r') as f:
            decoded = bencode.bdecode(f.read())

        cleaned = self.__remove_all_recursively(decoded)

        with open(self.output, 'w') as f:  # encode the decoded torrent
            f.write(bencode.bencode(cleaned))  # and write it in self.output

        self.do_backup()
        return True
